from pathlib import Path
from typing import Optional, List
from datetime import datetime
from dataclasses import dataclass
from itertools import islice
from functools import partial

from .files import DiagFile
from .nginx import try_fix_nginx_error


@dataclass
class LogFile(DiagFile):
    @classmethod
    def open(
        cls, path: Path, *, fix: bool = True, nginx_format_str: Optional[str] = None
    ) -> Optional["LogFile"]:
        log_file = super().open(path)
        if not log_file:
            return None
        if fix:
            log_file.try_fix(nginx_format_str)
        return log_file

    @classmethod
    def open_like(cls, path: Path, **kwargs) -> Optional["LogFile"]:
        name = path.name
        paths = find_other_files(path) + [path]
        log_files = [LogFile.open(path, **kwargs) for path in paths]
        log_files = list(filter(None, log_files))
        if not log_files:
            return None
        return LogFile.combine(log_files, name)

    def filter_by_datetime(self, start_date: datetime, end_date: datetime):
        self.lines = [
            line for line in self.lines if start_date <= parse_time(line) <= end_date
        ]

    def sort_by_datetime(self):
        self.lines.sort(key=parse_time)

    def try_fix(self: "LogFile", nginx_format_str: Optional[str] = None):
        self.map_lines(partial(try_fix_nginx_error, format_str=nginx_format_str))

        # sometimes logs contain a "\n",
        # which spreads the log record on multpile lines.
        # this function tries to find those lines
        # and merges them together
        prev_valid_idx = 0
        for i, line in islice(enumerate(self.lines, 0), 1, None):
            # TODO: use better validation
            if parse_time(line) != datetime.max:
                # line valid, carry on
                prev_valid_idx = i
            else:
                # line invalid, merge with last valid line
                prev_line: str = self.lines[prev_valid_idx]
                if prev_line.endswith("\n"):
                    prev_line = prev_line[:-1]
                prev_line += "\\n"
                prev_line += line
                self.lines[prev_valid_idx] = prev_line
                self.lines[i] = ""

        # filter out empty lines
        self.lines = list(filter(None, self.lines))

    @classmethod
    def combine(
        cls, log_files: List["LogFile"], name: str = "combined.log"
    ) -> "LogFile":
        lines: List[str] = sum((file.lines for file in log_files), [])
        # sort by time to retain chronology
        log_file = LogFile(name, lines)
        log_file.sort_by_datetime()
        return log_file


def parse_time(log_line: str) -> datetime:
    try:
        # 2023-12-07 15:54:24 | 1701964464956 | sdbapp | ...
        time_msecs = int(log_line.split(" | ")[1]) / 1000.0
        return datetime.fromtimestamp(time_msecs)
    except Exception:
        try:
            # 2023-12-12 15:04:05 1702393445834 [sdbapp] ...
            time_msecs = int(log_line.split()[2]) / 1000.0
            return datetime.fromtimestamp(time_msecs)
        except Exception:
            try:
                # 2023-12-12 15:04:04 - ...
                date_time_str = "T".join(log_line.split()[:2])
                return datetime.strptime(date_time_str, "%Y-%m-%dT%H:%M:%S")
            except Exception:
                return datetime.max


def find_other_files(log_file_path: Path) -> List[Path]:
    # log files for python and uwsgi are in format "{name}.log.{n}"
    # where n is either an increasing number or a timestamp
    # so this function matches those files so that they can be all combined
    # into a single file
    # it's not precise, but should be precise enough
    paths = []
    for file_path in log_file_path.parent.glob(f"{log_file_path.name}.*"):
        paths.append(file_path)
    return paths
